# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import queue
import asyncio
import threading
from functools import wraps

import inspect
from typing import Any, Callable, Coroutine, Awaitable, TypeVar, ParamSpec, Generator, AsyncGenerator
from asyncio import Future

T = TypeVar("T")
R = TypeVar("R")
P = ParamSpec("P")


class FunctionShifter:
    _future_loop = None
    _future_thread = None
    _future_lock = threading.Lock()

    @staticmethod
    def run_async_func_in_thread(func, *args, **kwargs):
        result: dict[str, Any] = {}

        def runner():
            try:
                result["data"] = asyncio.run(func(*args, **kwargs))
            except Exception as e:
                result["exception"] = e

        thread = threading.Thread(target=runner)
        thread.start()
        thread.join()

        if "exception" in result:
            raise result["exception"]
        return result["data"]

    @staticmethod
    def syncify(func: Callable[P, R | Awaitable[R]]) -> Callable[P, R]:
        if inspect.iscoroutinefunction(func):

            @wraps(func)
            def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
                try:
                    loop = asyncio.get_running_loop()
                except RuntimeError:
                    # No running loop, safe to use asyncio.run
                    return asyncio.run(func(*args, **kwargs))
                else:
                    # Running loop exists, move coroutine to thread
                    return FunctionShifter.run_async_func_in_thread(func, *args, **kwargs)

            return wrapper
        else:
            assert inspect.isfunction(func)
            return func

    @staticmethod
    def asyncify(func: Callable[P, R | Awaitable[R]]) -> Callable[P, Coroutine[Any, Any, R]]:
        if inspect.iscoroutinefunction(func):
            return func

        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            if not callable(func):
                raise TypeError(f"Expected a callable, got {type(func)}")
            return await asyncio.to_thread(func, *args, **kwargs)  # type: ignore

        return wrapper

    @staticmethod
    def future(func: Callable[P, R | Awaitable[R]]) -> Callable[P, Future[R]]:
        async_func = FunctionShifter.asyncify(func)

        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> Future[R]:
            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                if FunctionShifter._future_thread is None or FunctionShifter._future_loop is None:
                    with FunctionShifter._future_lock:
                        if FunctionShifter._future_thread is None or FunctionShifter._future_loop is None:
                            FunctionShifter._future_loop = asyncio.new_event_loop()
                            FunctionShifter._future_thread = threading.Thread(
                                target=FunctionShifter._future_loop.run_forever, daemon=True
                            ).start()
                loop = FunctionShifter._future_loop

            future = asyncio.ensure_future(async_func(*args, **kwargs), loop=loop)
            exception = future.add_done_callback(lambda t: t.exception())
            if exception:
                raise exception
            return future

        return wrapper

    @staticmethod
    def syncify_async_generator(async_gen: AsyncGenerator[R, Any]) -> Generator[R, Any, Any]:

        q: "queue.Queue" = queue.Queue()
        SENTINEL = object()

        async def _consume():
            try:
                async for item in async_gen:
                    q.put(("item", item))
            except Exception as e:
                q.put(("error", e))
            finally:
                q.put(("end", SENTINEL))

        def _runner():
            loop = asyncio.new_event_loop()
            try:
                asyncio.set_event_loop(loop)
                loop.run_until_complete(_consume())
            finally:
                loop.close()
                asyncio.set_event_loop(None)

        thread = threading.Thread(target=_runner, daemon=True)
        thread.start()

        try:
            while True:
                kind, payload = q.get()
                if kind == "item":
                    yield payload
                elif kind == "error":
                    raise payload
                elif kind == "end":
                    break
        finally:
            thread.join()

    @staticmethod
    def asyncify_sync_generator(sync_gen: Generator[R, Any, Any]) -> AsyncGenerator[R, Any]:
        SENTINEL = object()

        async def _wrapper():
            q: "asyncio.Queue[tuple[str, Any]]" = asyncio.Queue()
            loop = asyncio.get_running_loop()
            stop_event = threading.Event()

            def _put(kind: str, payload: Any):
                try:
                    loop.call_soon_threadsafe(q.put_nowait, (kind, payload))
                except RuntimeError:
                    pass

            def _consume():
                try:
                    while not stop_event.is_set():
                        item = next(sync_gen)
                        _put("item", item)
                except StopIteration:
                    pass
                except Exception as e:
                    _put("error", e)
                finally:
                    try:
                        sync_gen.close()
                    except Exception:
                        pass
                    _put("end", SENTINEL)

            thread = threading.Thread(target=_consume, daemon=True)
            thread.start()

            try:
                while True:
                    kind, payload = await q.get()
                    if kind == "item":
                        yield payload
                    elif kind == "error":
                        raise payload
                    elif kind == "end":
                        break
            finally:
                stop_event.set()
                await asyncio.to_thread(thread.join)

        return _wrapper()

    @staticmethod
    def auto_options_func(func: Callable[..., R]) -> Callable[..., R]:
        signature = inspect.signature(func)
        params = signature.parameters
        accepts_kwargs = any(p.kind == inspect.Parameter.VAR_KEYWORD for p in params.values())

        @wraps(func)
        def _wrapper(*args, **kwargs):
            if not accepts_kwargs:
                kwargs = {
                    name: value
                    for name, value in kwargs.items()
                    if name in params
                    and params[name].kind
                    in (
                        inspect.Parameter.POSITIONAL_OR_KEYWORD,
                        inspect.Parameter.KEYWORD_ONLY,
                    )
                }

            result = func(*args, **kwargs)
            return result

        return _wrapper



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .Logger import create_logger
from .Messenger import create_messenger
from .RuntimeData import RuntimeData, RuntimeDataNamespace
from .SerializableRuntimeData import SerializableRuntimeData, SerializableRuntimeDataNamespace
from .Settings import Settings, SettingsNamespace

# from .Storage import Storage, AsyncStorage
from .FunctionShifter import FunctionShifter
from .DataFormatter import DataFormatter
from .DataPathBuilder import DataPathBuilder
from .LazyImport import LazyImport
from .DataLocator import DataLocator
from .GeneratorConsumer import GeneratorConsumer
from .StreamingJSONCompleter import StreamingJSONCompleter
from .StreamingJSONParser import StreamingJSONParser
from .PythonSandbox import PythonSandbox
from .TimeInfo import TimeInfo



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import uuid
import asyncio
import warnings
from contextvars import ContextVar

from typing import Any, Literal, TYPE_CHECKING

if TYPE_CHECKING:
    from .TriggerFlow import TriggerFlow
    from agently.types.trigger_flow import TriggerFlowAllHandlers
    from agently.types.data import SerializableValue

from agently.utils import RuntimeData, FunctionShifter, GeneratorConsumer, Settings
from agently.types.trigger_flow import TriggerFlowEventData, RUNTIME_STREAM_STOP
from agently.types.data import EMPTY


class TriggerFlowExecution:
    def __init__(
        self,
        *,
        handlers: "TriggerFlowAllHandlers",
        trigger_flow: "TriggerFlow",
        id: str | None = None,
        skip_exceptions: bool = False,
        concurrency: int | None = None,
    ):
        # Basic Attributions
        self.id = id if id is not None else uuid.uuid4().hex
        self._handlers = handlers
        self._trigger_flow = trigger_flow
        self._runtime_data = RuntimeData()
        self._system_runtime_data = RuntimeData()
        self._skip_exceptions = skip_exceptions
        self._concurrency_semaphore = asyncio.Semaphore(concurrency) if concurrency and concurrency > 0 else None
        self._concurrency_depth = ContextVar(
            f"trigger_flow_execution_concurrency_depth_{ self.id }",
            default=0,
        )

        # Settings
        self.settings = Settings(
            parent=self._trigger_flow.settings,
            name=f"TriggerFlowExecution-{ self.id }-Settings",
        )

        # Emit
        self.emit = FunctionShifter.syncify(self.async_emit)

        # Flow Data
        self.get_flow_data = self._trigger_flow.get_flow_data
        self.set_flow_data = self._trigger_flow.set_flow_data
        self.async_set_flow_data = self._trigger_flow.async_set_flow_data
        self.append_flow_data = self._trigger_flow.append_flow_data
        self.async_append_flow_data = self._trigger_flow.async_append_flow_data
        self.del_flow_data = self._trigger_flow.del_flow_data
        self.async_del_flow_data = self._trigger_flow.async_del_flow_data

        # Runtime Data
        self.get_runtime_data = self._runtime_data.get
        self.set_runtime_data = FunctionShifter.syncify(self.async_set_runtime_data)
        self.append_runtime_data = FunctionShifter.syncify(self.async_append_runtime_data)
        self.del_runtime_data = FunctionShifter.syncify(self.async_del_runtime_data)

        # Start
        self.start = FunctionShifter.syncify(self.async_start)

        # Runtime Stream
        self.put_into_stream = FunctionShifter.syncify(self.async_put_into_stream)
        self.stop_stream = FunctionShifter.syncify(self.async_stop_stream)

        # Result
        self.get_result = FunctionShifter.syncify(self.async_get_result)

        # Execution Status
        self._started = False
        self._system_runtime_data.set("result", EMPTY)
        self._system_runtime_data.set("result_ready", asyncio.Event())
        self._runtime_stream_queue = asyncio.Queue()
        self._runtime_stream_consumer: GeneratorConsumer | None = None

    # Set Settings
    def set_settings(self, key: str, value: "SerializableValue"):
        self.settings.set_settings(key, value)
        return self

    # Set Concurrency
    def set_concurrency(self, concurrency):
        self._concurrency_semaphore = asyncio.Semaphore(concurrency) if concurrency and concurrency > 0 else None
        return self

    # Emit Event
    async def async_emit(
        self,
        trigger_event: str,
        value: Any = None,
        _layer_marks: list[str] | None = None,
        *,
        trigger_type: Literal["event", "runtime_data", "flow_data"] = "event",
    ):
        from agently.base import async_system_message

        await async_system_message(
            "TRIGGER_FLOW",
            {
                "TYPE": trigger_type,
                "EVENT": trigger_event,
                "VALUE": value,
            },
            self.settings,
        )
        tasks = []
        handlers = self._handlers[trigger_type]

        if trigger_event in handlers:
            for handler_id, handler in handlers[trigger_event].items():
                await async_system_message(
                    "TRIGGER_FLOW",
                    {
                        "EVENT": trigger_event,
                        "TYPE": trigger_type,
                        "HANDLER": handler_id,
                    },
                    self.settings,
                )

                async def run_handler(handler_func):
                    if self._concurrency_semaphore is None:
                        return await handler_func
                    depth = self._concurrency_depth.get()
                    token = self._concurrency_depth.set(depth + 1)
                    try:
                        if depth > 0:
                            return await handler_func
                        async with self._concurrency_semaphore:
                            return await handler_func
                    finally:
                        self._concurrency_depth.reset(token)

                handler_task = FunctionShifter.asyncify(handler)(
                    TriggerFlowEventData(
                        trigger_event=trigger_event,
                        trigger_type=trigger_type,
                        value=value,
                        execution=self,
                        _layer_marks=_layer_marks,
                    )
                )
                tasks.append(asyncio.ensure_future(run_handler(handler_task)))

        if tasks:
            await asyncio.gather(*tasks, return_exceptions=self._skip_exceptions)

    # Change Runtime Data
    async def _async_change_runtime_data(
        self,
        operation: Literal["set", "append", "del"],
        key: str,
        value: Any,
        *,
        emit: bool = True,
    ):
        futures = []
        handlers = self._handlers["runtime_data"]

        match operation:
            case "set":
                self._runtime_data.set(key, value)
                value = self._runtime_data[key]
            case "append":
                self._runtime_data.append(key, value)
                value = self._runtime_data[key]
            case "del":
                if self._runtime_data.get(key, None):
                    del self._runtime_data[key]
                    value = None
                else:
                    return
        if emit:
            if key in handlers:
                futures.append(
                    self.async_emit(
                        key,
                        value,
                        trigger_type="runtime_data",
                    )
                )

            if futures:
                await asyncio.gather(*futures, return_exceptions=self._skip_exceptions)

    async def async_set_runtime_data(
        self,
        key: str,
        value: Any,
        *,
        emit: bool = True,
    ):
        return await self._async_change_runtime_data("set", key, value, emit=emit)

    async def async_append_runtime_data(
        self,
        key: str,
        value: Any,
        *,
        emit: bool = True,
    ):
        return await self._async_change_runtime_data("append", key, value, emit=emit)

    async def async_del_runtime_data(
        self,
        key: str,
        *,
        emit: bool = True,
    ):
        return await self._async_change_runtime_data("del", key, None, emit=emit)

    # Start
    async def async_start(
        self,
        initial_value: Any = None,
        *,
        wait_for_result: bool = True,
        timeout: float | None = 10,
    ):
        await self.async_emit("START", initial_value)
        if wait_for_result:
            return await self.async_get_result(timeout=timeout)

    # Runtime Stream
    async def async_put_into_stream(self, stream_item: Any):
        await self._runtime_stream_queue.put(stream_item)

    async def async_stop_stream(self):
        await self._runtime_stream_queue.put(RUNTIME_STREAM_STOP)

    async def _consume_runtime_stream(
        self,
        *,
        initial_value: Any,
        timeout: float | None,
    ):
        temp_execution_task = None
        try:
            if not self._started:
                temp_execution_task = asyncio.create_task(
                    self.async_start(
                        initial_value=initial_value,
                        wait_for_result=False,
                    )
                )
            while True:
                if timeout is not None:
                    try:
                        next_result = await asyncio.wait_for(
                            self._runtime_stream_queue.get(),
                            timeout=timeout,
                        )
                    except asyncio.TimeoutError:
                        warnings.warn(
                            f"Execution { self.id } runtime stream stopped because of timeout.\n"
                            f"Timeout seconds: { timeout }\n"
                            "You can use execution.get_async_runtime_stream(timeout=<int | None>) or execution.get_runtime_stream(timeout=<int | None>) to reset new timeout seconds or use None to wait forever."
                        )
                        break
                else:
                    next_result = await self._runtime_stream_queue.get()
                if next_result is not RUNTIME_STREAM_STOP:
                    yield next_result
                else:
                    break
        finally:
            if temp_execution_task:
                await temp_execution_task

    def get_async_runtime_stream(
        self,
        initial_value: Any = None,
        *,
        timeout: float | None = 10,
    ):
        if self._runtime_stream_consumer is None:
            self._runtime_stream_consumer = GeneratorConsumer(
                self._consume_runtime_stream(
                    initial_value=initial_value,
                    timeout=timeout,
                )
            )
        return self._runtime_stream_consumer.get_async_generator()

    def get_runtime_stream(
        self,
        initial_value: Any = None,
        *,
        timeout: float | None = 10,
    ):
        if self._runtime_stream_consumer is None:
            self._runtime_stream_consumer = GeneratorConsumer(
                self._consume_runtime_stream(
                    initial_value=initial_value,
                    timeout=timeout,
                )
            )
        return self._runtime_stream_consumer.get_generator()

    # Result
    def set_result(self, result: Any):
        self._system_runtime_data.set("result", result)
        result_ready = self._system_runtime_data.get("result_ready")
        if isinstance(result_ready, asyncio.Event):
            result_ready.set()

    async def async_get_result(self, *, timeout: float | None = None):
        if timeout is None:
            result_ready = self._system_runtime_data.get("result_ready")
            if isinstance(result_ready, asyncio.Event):
                await result_ready.wait()
            self._result = self._system_runtime_data.get("result")
            return self._result
        else:
            try:
                result_ready = self._system_runtime_data.get("result_ready")
                if isinstance(result_ready, asyncio.Event):
                    await asyncio.wait_for(result_ready.wait(), timeout=timeout)
                self._result = self._system_runtime_data.get("result")
                return self._result
            except asyncio.TimeoutError:
                warnings.warn(
                    f"Can not get the result of trigger flow { self.id } for it took too long and timeout.\n"
                    "You can check if you forget to use flow.set_result() to set a result for this trigger flow. Or you can set parameter 'timeout' to a bigger number to wait longer or to None to wait forever."
                    f"Timeout: { timeout }"
                )
                self._result = None
                return self._result



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio

import json
from typing import TYPE_CHECKING, Any

from agently.types.data import SerializableData, EventMessage
from agently.utils import FunctionShifter

if TYPE_CHECKING:
    from agently.types.data import (
        AgentlyEvent,
        AgentlySystemEvent,
        EventMessageDict,
        EventMessage,
        MessageLevel,
        EventStatus,
        EventHook,
    )
    from agently.types.plugins import EventHooker
    from agently.utils import Settings


class EventCenter:
    def __init__(self):
        self._hooks: dict[AgentlyEvent, dict[str, "EventHook"]] = {}
        self._hookers: dict[str, type[EventHooker]] = {}
        self.emit = FunctionShifter.syncify(self.async_emit)
        self.system_message = FunctionShifter.syncify(self.async_system_message)

    def register_hook(
        self,
        event: "AgentlyEvent",
        callback: "EventHook",
        *,
        hook_name: str | None = None,
    ):
        if hook_name is None:
            hook_name = callback.__name__
        if event not in self._hooks:
            self._hooks.update({event: {}})
        self._hooks[event].update({hook_name: callback})

    def unregister_hook(
        self,
        event: "AgentlyEvent",
        hook_name: str,
    ):
        if event in self._hooks and hook_name in self._hooks[event]:
            del self._hooks[event][hook_name]

    def register_hooker_plugin(self, hooker: type["EventHooker"]):
        if hasattr(hooker, "_on_register"):
            hooker._on_register()
        for event in hooker.events:
            self.register_hook(event, hooker.handler, hook_name=hooker.name)
        self._hookers.update({hooker.name: hooker})

    def unregister_hooker_plugin(self, hooker: str | type["EventHooker"]):
        if isinstance(hooker, str):
            if hooker in self._hookers:
                hooker = self._hookers[hooker]
            else:
                return
        for event in hooker.events:
            self.unregister_hook(event, hook_name=hooker.name)
        if hasattr(hooker, "_on_unregister"):
            hooker._on_unregister()
        del self._hookers[hooker.name]

    async def async_emit(
        self,
        event: "AgentlyEvent",
        message: "EventMessageDict | EventMessage",
    ):
        if message is EventMessage:
            message_object = message
        else:
            message_dict = dict(message).copy()
            message_dict.update({"event": event})
            message_object = EventMessage(**message_dict)

        tasks = []

        if event in self._hooks:
            for callback in self._hooks[event].values():
                coro = FunctionShifter.asyncify(callback)
                tasks.append(
                    asyncio.create_task(coro(message_object)),
                )
            await asyncio.gather(*tasks, return_exceptions=True)
        if event == "log" and len(tasks) == 0:
            print(*message_object.content if isinstance(message_object.content, list) else message_object.content)

    async def async_system_message(
        self,
        message_type: "AgentlySystemEvent",
        message_data: Any,
        settings: "Settings | None" = None,
    ):
        if settings is None:
            from agently.base import settings as default_settings

            settings = default_settings

        await self.async_emit(
            "AGENTLY_SYS",
            {
                "module_name": "Agently",
                "content": {
                    "type": message_type,
                    "data": message_data,
                    "settings": settings,
                },
            },
        )

    def create_messenger(self, module_name: str, *, base_meta: dict[str, Any] | None = None):
        if base_meta is None:
            base_meta = {}
        return EventCenterMessenger(
            self,
            module_name,
            base_meta=base_meta,
        )


class EventCenterMessenger:
    def __init__(
        self,
        event_center: EventCenter,
        module_name: str,
        *,
        base_meta: dict[str, Any] | None = None,
    ):
        self._event_center = event_center
        self._module_name = module_name
        self._base_meta = base_meta if base_meta is not None else {}

        self.message = FunctionShifter.syncify(self.async_message)
        self.debug = FunctionShifter.syncify(self.async_debug)
        self.info = FunctionShifter.syncify(self.async_info)
        self.warning = FunctionShifter.syncify(self.async_warning)
        self.error = FunctionShifter.syncify(self.async_error)
        self.critical = FunctionShifter.syncify(self.async_critical)
        self.to_console = FunctionShifter.syncify(self.async_to_console)
        self.to_data = FunctionShifter.syncify(self.async_to_data)

    def update_base_meta(self, update_dict: dict[str, Any]):
        self._base_meta.update(update_dict)

    async def async_message(
        self,
        content: str,
        *,
        event: "AgentlyEvent" = "message",
        status: "EventStatus" = "",
        exception: Exception | None = None,
        level: "MessageLevel" = "INFO",
        meta: dict[str, Any] | None = None,
    ):
        if meta is None:
            meta = {}
        final_meta = self._base_meta.copy()
        final_meta.update(meta)
        await self._event_center.async_emit(
            event,
            {
                "module_name": self._module_name,
                "status": status,
                "content": content,
                "exception": exception,
                "level": level,
                "meta": final_meta,
            },
        )

    async def async_debug(
        self,
        content: str,
        *,
        status: "EventStatus" = "",
        meta: dict[str, Any] | None = None,
    ):
        if meta is None:
            meta = {}
        final_meta = self._base_meta.copy()
        final_meta.update(meta)
        await self._event_center.async_emit(
            "log",
            {
                "module_name": self._module_name,
                "status": status,
                "content": content,
                "level": "DEBUG",
                "meta": final_meta,
            },
        )

    async def async_info(
        self,
        content: str,
        *,
        status: "EventStatus" = "",
        meta: dict[str, Any] | None = None,
    ):
        if meta is None:
            meta = {}
        final_meta = self._base_meta.copy()
        final_meta.update(meta)
        await self._event_center.async_emit(
            "log",
            {
                "module_name": self._module_name,
                "status": status,
                "content": content,
                "level": "INFO",
                "meta": final_meta,
            },
        )

    async def async_warning(
        self,
        content: str,
        *,
        status: "EventStatus" = "",
        meta: dict[str, Any] | None = None,
    ):
        if meta is None:
            meta = {}
        final_meta = self._base_meta.copy()
        final_meta.update(meta)
        await self._event_center.async_emit(
            "log",
            {
                "module_name": self._module_name,
                "status": status,
                "content": content,
                "level": "WARNING",
                "meta": final_meta,
            },
        )

    async def async_error(
        self,
        error: str | Exception,
        *,
        status: "EventStatus" = "",
        meta: dict[str, Any] | None = None,
    ):
        if meta is None:
            meta = {}
        final_meta = self._base_meta.copy()
        final_meta.update(meta)
        error = error if isinstance(error, Exception) else RuntimeError(error)
        await self._event_center.async_emit(
            "log",
            {
                "module_name": self._module_name,
                "status": status,
                "content": str(error),
                "exception": error,
                "level": "ERROR",
                "meta": final_meta,
            },
        )

        from agently.base import settings

        if settings.get("runtime.raise_error"):
            raise error

    async def async_critical(
        self,
        critical: str | Exception,
        *,
        status: "EventStatus" = "",
        meta: dict[str, Any] | None = None,
    ):
        if meta is None:
            meta = {}
        final_meta = self._base_meta.copy()
        final_meta.update(meta)
        critical = critical if isinstance(critical, Exception) else RuntimeError(critical)
        await self._event_center.async_emit(
            "log",
            {
                "module_name": self._module_name,
                "status": status,
                "content": str(critical),
                "exception": critical,
                "level": "CRITICAL",
                "meta": final_meta,
            },
        )

        from agently.base import settings

        if settings.get("runtime.raise_critical"):
            raise critical

    async def async_to_console(
        self,
        content: Any,
        *,
        status: "EventStatus" = "",
        table_name: str | None = None,
        row_id: int | str | None = None,
    ):
        final_meta = {}
        final_meta.update(self._base_meta)
        if table_name is not None:
            final_meta.update({"table_name": table_name})
        if row_id is not None:
            final_meta.update({"row_id": row_id})
        await self._event_center.async_emit(
            "console",
            {
                "module_name": self._module_name,
                "status": status,
                "content": content,
                "meta": final_meta,
            },
        )

    async def async_to_data(
        self,
        data: SerializableData,
        *,
        status: "EventStatus" = "",
        meta: dict[str, Any],
    ):
        final_meta = self._base_meta
        final_meta.update(meta)
        await self._event_center.async_emit(
            "data",
            {
                "module_name": self._module_name,
                "status": status,
                "content": json.dumps(data),
                "meta": final_meta,
            },
        )



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import json
import uuid

from typing import AsyncGenerator, TYPE_CHECKING, cast

from agently.core.Prompt import Prompt
from agently.core.ExtensionHandlers import ExtensionHandlers
from agently.utils import Settings, DataFormatter

from agently.core.ModelResponseResult import ModelResponseResult

if TYPE_CHECKING:
    from agently.core import PluginManager
    from agently.types.data import AgentlyModelResponseMessage
    from agently.types.plugins import ModelRequester


class ModelResponse:
    def __init__(
        self,
        agent_name: str,
        plugin_manager: "PluginManager",
        settings: Settings,
        prompt: Prompt,
        extension_handlers: ExtensionHandlers,
    ):
        self.agent_name = agent_name
        self.id = uuid.uuid4().hex
        self.plugin_manager = plugin_manager
        settings_snapshot = settings.get()
        self.settings = Settings(settings_snapshot if isinstance(settings_snapshot, dict) else {})
        self.settings.set("$log.cancel_logs", False)
        prompt_snapshot = prompt.get()
        self.prompt = Prompt(
            self.plugin_manager,
            self.settings,
            prompt_dict=prompt_snapshot if isinstance(prompt_snapshot, dict) else {},
        )
        extension_handlers_snapshot = extension_handlers.get()
        self.extension_handlers = ExtensionHandlers(
            extension_handlers_snapshot if isinstance(extension_handlers_snapshot, dict) else {}
        )
        self.result = ModelResponseResult(
            self.agent_name,
            self.id,
            self.prompt,
            self._get_response_generator(),
            self.plugin_manager,
            self.settings,
            self.extension_handlers,
        )
        self.get_meta = self.result.get_meta
        self.async_get_meta = self.result.async_get_meta
        self.get_text = self.result.get_text
        self.async_get_text = self.result.async_get_text
        self.get_data = self.result.get_data
        self.async_get_data = self.result.async_get_data
        self.get_data_object = self.result.get_data_object
        self.async_get_data_object = self.result.async_get_data_object
        self.get_generator = self.result.get_generator
        self.get_async_generator = self.result.get_async_generator

    def cancel_logs(self):
        self.settings.set("$log.cancel_logs", True)

    async def _get_response_generator(self) -> AsyncGenerator["AgentlyModelResponseMessage", None]:
        from agently.base import async_system_message

        ModelRequester = cast(
            type["ModelRequester"],
            self.plugin_manager.get_plugin(
                "ModelRequester",
                str(self.settings["plugins.ModelRequester.activate"]),
            ),
        )
        request_prefixes = self.extension_handlers.get("request_prefixes", [])
        for prefix in request_prefixes:
            if inspect.iscoroutinefunction(prefix):
                await prefix(self.prompt, self.settings)
            elif inspect.isfunction(prefix):
                prefix(self.prompt, self.settings)
        model_requester = ModelRequester(self.prompt, self.settings)
        request_data = model_requester.generate_request_data()
        request_data_dict = DataFormatter.sanitize(request_data.model_dump())
        await async_system_message(
            "MODEL_REQUEST",
            {
                "agent_name": self.agent_name,
                "response_id": self.id,
                "content": {
                    "stage": "Requesting",
                    "detail": json.dumps(
                        {
                            "data": request_data_dict["data"] if "data" in request_data_dict else None,
                            "request_options": (
                                request_data_dict["request_options"] if "request_options" in request_data_dict else None
                            ),
                            "request_url": (
                                request_data_dict["request_url"] if "request_url" in request_data_dict else None
                            ),
                            "stream": (request_data_dict["stream"] if "stream" in request_data_dict else None),
                        },
                        indent=2,
                        ensure_ascii=False,
                    )
                    .replace("\\n", "\n")
                    .replace("\\\"", "\""),
                },
            },
            self.settings,
        )
        response_generator = model_requester.request_model(request_data)
        broadcast_generator = model_requester.broadcast_response(response_generator)
        broadcast_prefixes = self.extension_handlers.get("broadcast_prefixes", [])
        broadcast_suffixes = self.extension_handlers.get("broadcast_suffixes", {})
        for prefix in broadcast_prefixes:
            if inspect.iscoroutinefunction(prefix):
                result = await prefix(
                    self.result.full_result_data,
                    self.settings,
                )
                if result is not None:
                    yield result
            elif inspect.isgeneratorfunction(prefix):
                for result in prefix(
                    self.result.full_result_data,
                    self.settings,
                ):
                    if result is not None:
                        yield result
            elif inspect.isasyncgenfunction(prefix):
                async for result in prefix(
                    self.result.full_result_data,
                    self.settings,
                ):
                    if result is not None:
                        yield result
            elif inspect.isfunction(prefix):
                result = prefix(
                    self.result.full_result_data,
                    self.settings,
                )
                if result is not None:
                    yield result
        async for event, data in broadcast_generator:
            yield event, data
            suffixes = broadcast_suffixes[event] if event in broadcast_suffixes else []
            for suffix in suffixes:
                if inspect.iscoroutinefunction(suffix):
                    result = await suffix(
                        event,
                        data,
                        self.result.full_result_data,
                        self.settings,
                    )
                    if result is not None:
                        yield result
                elif inspect.isgeneratorfunction(suffix):
                    for result in suffix(
                        event,
                        data,
                        self.result.full_result_data,
                        self.settings,
                    ):
                        if result is not None:
                            yield result
                elif inspect.isasyncgenfunction(suffix):
                    async for result in suffix(
                        event,
                        data,
                        self.result.full_result_data,
                        self.settings,
                    ):
                        if result is not None:
                            yield result
                elif inspect.isfunction(suffix):
                    result = suffix(
                        event,
                        data,
                        self.result.full_result_data,
                        self.settings,
                    )
                    if result is not None:
                        yield result



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

from typing import Any, Callable, Literal, TYPE_CHECKING
from agently.types.data import AVOID_COPY

if TYPE_CHECKING:
    from agently.core import TriggerFlowExecution

from agently.utils import RuntimeData


class TriggerFlowBlockData:
    global_data = RuntimeData()

    def __init__(
        self,
        outer_block: "TriggerFlowBlockData | None" = None,
        data: dict[str, Any] | None = None,
    ):
        self.outer_block = outer_block
        self.data = data if data is not None else {}


class TriggerFlowEventData:
    def __init__(
        self,
        *,
        trigger_event: str,
        trigger_type: Literal["event", "runtime_data", "flow_data"],
        value: Any,
        execution: "TriggerFlowExecution",
        _layer_marks: list[str] | None = None,
    ):
        self.trigger_event = trigger_event
        self.trigger_type = trigger_type
        self.event = trigger_event
        self.type = trigger_type
        self.value = value
        self.execution_id = execution.id
        self._layer_marks = _layer_marks if _layer_marks is not None else []
        self.settings = execution.settings

        self.get_flow_data = execution.get_flow_data
        self.set_flow_data = execution.set_flow_data
        self.append_flow_data = execution.append_flow_data
        self.del_flow_data = execution.del_flow_data
        self.async_set_flow_data = execution.async_set_flow_data
        self.async_append_flow_data = execution.async_append_flow_data
        self.async_del_flow_data = execution.async_del_flow_data

        self.get_runtime_data = execution.get_runtime_data
        self.set_runtime_data = execution.set_runtime_data
        self.append_runtime_data = execution.append_runtime_data
        self.del_runtime_data = execution.del_runtime_data
        self.async_set_runtime_data = execution.async_set_runtime_data
        self.async_append_runtime_data = execution.async_append_runtime_data
        self.async_del_runtime_data = execution.async_del_runtime_data

        self.emit = execution.emit
        self.async_emit = execution.async_emit

        self.put = execution.put_into_stream
        self.async_put = execution.async_put_into_stream
        self.put_into_stream = execution.put_into_stream
        self.async_put_into_stream = execution.async_put_into_stream
        self.stop_stream = execution.stop_stream
        self.async_stop_stream = execution.async_stop_stream

        self._system_runtime_data = execution._system_runtime_data

    @property
    def upper_layer_mark(self):
        return self._layer_marks[-2] if len(self._layer_marks) > 1 else None

    @property
    def layer_mark(self):
        return self._layer_marks[-1] if len(self._layer_marks) > 0 else None

    def layer_in(self):
        self._layer_marks.append(uuid.uuid4().hex)

    def layer_out(self):
        self._layer_marks = self._layer_marks[:-1] if len(self._layer_marks) > 0 else []


TriggerFlowHandler = Callable[[TriggerFlowEventData], Any]
TriggerFlowHandlers = dict[str, dict[str, TriggerFlowHandler]]
TriggerFlowAllHandlers = dict[Literal["event", "flow_data", "runtime_data"], TriggerFlowHandlers]

RUNTIME_STREAM_STOP = AVOID_COPY()



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Any, Literal, Type, TYPE_CHECKING, TypeVar, Generic, cast

from agently.utils import Settings, create_logger, FunctionShifter, DataFormatter
from agently.core import PluginManager, EventCenter, Tool, Prompt, ModelRequest, BaseAgent
from agently._default_init import _load_default_settings, _load_default_plugins, _hook_default_event_handlers

if TYPE_CHECKING:
    from agently.types.data import MessageLevel, SerializableValue

# Basic Initialize

settings = Settings(
    name="global_settings",
)
_load_default_settings(settings)
plugin_manager = PluginManager(
    settings,
    name="global_plugin_manager",
)
_load_default_plugins(plugin_manager)
event_center = EventCenter()
_hook_default_event_handlers(event_center)
async_system_message = event_center.async_system_message
system_message = event_center.system_message
logger = create_logger()
httpx_level_name = settings.get("runtime.httpx_log_level", "WARNING")
httpx_level = getattr(logging, str(httpx_level_name).upper(), logging.WARNING)
logging.getLogger("httpx").setLevel(httpx_level)
logging.getLogger("httpcore").setLevel(httpx_level)
tool = Tool(plugin_manager, settings)
_agently_messenger = event_center.create_messenger("Agently")


def print_(content: Any, *args):
    message_sync = FunctionShifter.syncify(_agently_messenger.message)
    contents = [str(content)]
    if args:
        for arg in args:
            contents.append(str(arg))
    content_text = " ".join(contents)
    message_sync(content_text, event="log")


async def async_print(content: Any, *args):
    contents = [str(content)]
    if args:
        for arg in args:
            contents.append(str(arg))
    content_text = " ".join(contents)
    await _agently_messenger.async_message(content_text, event="log")


# Settings Mappings

settings.update_mappings(
    {
        "key_value_mappings": {
            "debug": {
                True: {
                    "runtime.show_model_logs": True,
                    "runtime.show_tool_logs": True,
                    "runtime.show_trigger_flow_logs": True,
                    "runtime.httpx_log_level": "INFO",
                },
                False: {
                    "runtime.show_model_logs": False,
                    "runtime.show_tool_logs": False,
                    "runtime.show_trigger_flow_logs": False,
                    "runtime.httpx_log_level": "WARNING",
                },
            }
        }
    }
)

# Extensions Installation
# BaseAgent + Extensions = Agent
from agently.builtins.agent_extensions import (
    StreamingPrintExtension,
    SessionExtension,
    ToolExtension,
    KeyWaiterExtension,
    AutoFuncExtension,
    ConfigurePromptExtension,
)


class Agent(
    StreamingPrintExtension,
    SessionExtension,
    ToolExtension,
    KeyWaiterExtension,
    AutoFuncExtension,
    ConfigurePromptExtension,
    BaseAgent,
): ...


A = TypeVar("A", bound=Agent)

# Agently Main


class AgentlyMain(Generic[A]):
    def __init__(self, AgentType: Type[A] = Agent):
        self.settings = settings
        self.plugin_manager = plugin_manager
        self.event_center = event_center
        self.logger = logger
        self.print = print_
        self.async_print = async_print
        self.tool = tool
        self.AgentType = AgentType

        def set_settings(key: str, value: "SerializableValue", *, auto_load_env: bool = False):
            self.settings.set_settings(key, value, auto_load_env=auto_load_env)
            if key in ("runtime.httpx_log_level", "debug"):
                level_name = self.settings.get("runtime.httpx_log_level", "WARNING")
                level = getattr(logging, str(level_name).upper(), logging.WARNING)
                logging.getLogger("httpx").setLevel(level)
                logging.getLogger("httpcore").setLevel(level)
            return self

        self.set_settings = set_settings

    def set_debug_console(self, debug_console_status: Literal["ON", "OFF"]):
        # Deprecated: debug console mode is retired and no longer participates in runtime.
        if debug_console_status == "ON":
            self.logger.warning("`set_debug_console(\"ON\")` is deprecated and has no effect.")
        return self

    def set_log_level(self, log_level: "MessageLevel"):
        self.logger.setLevel(log_level)
        return self

    def create_prompt(self, name: str = "agently_prompt") -> Prompt:
        return Prompt(
            self.plugin_manager,
            self.settings,
            name=name,
        )

    def create_request(self, name: str | None = None) -> ModelRequest:
        return ModelRequest(
            self.plugin_manager,
            parent_settings=self.settings,
            agent_name=name,
        )

    def create_agent(self, name: str | None = None) -> A:
        return self.AgentType(
            self.plugin_manager,
            parent_settings=self.settings,
            name=name,
        )



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import warnings

from typing import Any, Callable, TYPE_CHECKING, TypeVar, ParamSpec

from agently.utils import FunctionShifter
from agently.core import BaseAgent

if TYPE_CHECKING:
    from agently.core import Prompt
    from agently.core.Tool import ToolCommand, ToolExecutionRecord
    from agently.types.data import KwargsType, ReturnType, MCPConfigs, AgentlyModelResult

from agently.base import tool

P = ParamSpec("P")
R = TypeVar("R")


class ToolExtension(BaseAgent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from agently.base import async_system_message

        self.tool = tool
        self.tool_func = self.tool.tool_func

        self.use_tool = self.use_tools
        self.use_mcp = FunctionShifter.syncify(self.async_use_mcp)

        self.settings.setdefault("tool.loop.max_rounds", 5, inherit=True)
        self.settings.setdefault("tool.loop.concurrency", None, inherit=True)
        self.settings.setdefault("tool.loop.timeout", None, inherit=True)
        self.settings.setdefault("tool.loop.enabled", True, inherit=True)

        self.__tool_logs: list[ToolExecutionRecord] = []

        self.extension_handlers.append("request_prefixes", self.__request_prefix)
        self.extension_handlers.append("broadcast_prefixes", self.__broadcast_prefix)

        self.async_system_message = async_system_message

    def register_tool(
        self,
        *,
        name: str,
        desc: str,
        kwargs: "KwargsType",
        func: Callable,
        returns: "ReturnType | None" = None,
    ):
        self.tool.register(
            name=name,
            desc=desc,
            kwargs=kwargs,
            func=func,
            tags=[f"agent-{ self.name }"],
            returns=returns,
        )
        return self

    def tool_func(self, func: Callable[P, R]) -> Callable[P, R]:
        self.tool.tool_func(func)
        name = func.__name__
        self.tool.tag([name], [f"agent-{ self.name }"])
        return func

    def use_tools(self, tools: Callable | str | list[str | Callable]):
        if isinstance(tools, (str, Callable)):
            tools = [tools]
        names = []
        for tool in tools:
            if isinstance(tool, str):
                names.append(tool)
            else:
                if tool.__name__ not in self.tool.tool_manager.tool_funcs:  # type: ignore
                    self.tool_func(tool)
                names.append(tool.__name__)
        self.tool.tag(names, f"agent-{ self.name }")
        return self

    async def async_use_mcp(self, transport: "MCPConfigs | str | Any"):
        await self.tool.async_use_mcp(transport, tags=[f"agent-{ self.name }"])
        return self

    def set_tool_loop(
        self,
        *,
        enabled: bool | None = None,
        max_rounds: int | None = None,
        concurrency: int | None = None,
        timeout: float | None = None,
    ):
        if enabled is not None:
            self.settings.set("tool.loop.enabled", bool(enabled))
        if max_rounds is not None:
            if not isinstance(max_rounds, int) or max_rounds < 0:
                raise ValueError("max_rounds must be an integer >= 0.")
            self.settings.set("tool.loop.max_rounds", max_rounds)
        if concurrency is not None:
            if not isinstance(concurrency, int) or concurrency <= 0:
                raise ValueError("concurrency must be an integer > 0.")
            self.settings.set("tool.loop.concurrency", concurrency)
        if timeout is not None:
            if not isinstance(timeout, (int, float)) or timeout <= 0:
                raise ValueError("timeout must be a number > 0.")
            self.settings.set("tool.loop.timeout", float(timeout))
        return self

    def register_tool_plan_analysis_handler(self, handler):
        self.tool.register_plan_analysis_handler(handler)
        return self

    def register_tool_execution_handler(self, handler):
        self.tool.register_tool_execution_handler(handler)
        return self

    async def async_generate_tool_command(
        self,
        prompt: "Prompt | None" = None,
        *,
        done_plans: list["ToolExecutionRecord"] | None = None,
        last_round_records: list["ToolExecutionRecord"] | None = None,
        round_index: int = 0,
        max_rounds: int | None = None,
    ) -> list["ToolCommand"]:
        target_prompt = prompt if prompt is not None else self.request.prompt
        tool_list = self.tool.get_tool_list(tags=[f"agent-{ self.name }"])
        return await self.tool.async_generate_tool_command(
            prompt=target_prompt,
            settings=self.settings,
            tool_list=tool_list,
            agent_name=self.name,
            done_plans=done_plans,
            last_round_records=last_round_records,
            round_index=round_index,
            max_rounds=max_rounds,
        )

    def generate_tool_command(
        self,
        prompt: "Prompt | None" = None,
        *,
        done_plans: list["ToolExecutionRecord"] | None = None,
        last_round_records: list["ToolExecutionRecord"] | None = None,
        round_index: int = 0,
        max_rounds: int | None = None,
    ) -> list["ToolCommand"]:
        return FunctionShifter.syncify(self.async_generate_tool_command)(
            prompt=prompt,
            done_plans=done_plans,
            last_round_records=last_round_records,
            round_index=round_index,
            max_rounds=max_rounds,
        )

    async def async_must_call(
        self,
        prompt: "Prompt | None" = None,
        *,
        done_plans: list["ToolExecutionRecord"] | None = None,
        last_round_records: list["ToolExecutionRecord"] | None = None,
        round_index: int = 0,
        max_rounds: int | None = None,
    ) -> list["ToolCommand"]:
        warnings.warn(
            "Method .async_must_call() is deprecated and will be removed in future version, "
            "please use .async_generate_tool_command() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return await self.async_generate_tool_command(
            prompt=prompt,
            done_plans=done_plans,
            last_round_records=last_round_records,
            round_index=round_index,
            max_rounds=max_rounds,
        )

    def must_call(
        self,
        prompt: "Prompt | None" = None,
        *,
        done_plans: list["ToolExecutionRecord"] | None = None,
        last_round_records: list["ToolExecutionRecord"] | None = None,
        round_index: int = 0,
        max_rounds: int | None = None,
    ) -> list["ToolCommand"]:
        warnings.warn(
            "Method .must_call() is deprecated and will be removed in future version, "
            "please use .generate_tool_command() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.generate_tool_command(
            prompt=prompt,
            done_plans=done_plans,
            last_round_records=last_round_records,
            round_index=round_index,
            max_rounds=max_rounds,
        )

    async def __request_prefix(self, prompt: "Prompt", _):
        self.__tool_logs = []
        if self.settings.get("tool.loop.enabled", True) is not True:
            return

        tool_list = self.tool.get_tool_list(tags=[f"agent-{ self.name }"])
        if len(tool_list) == 0:
            return

        records = await self.tool.async_plan_and_execute(
            prompt=prompt,
            settings=self.settings,
            tool_list=tool_list,
            agent_name=self.name,
            max_rounds=self.settings.get("tool.loop.max_rounds", 5),  # type: ignore
            concurrency=self.settings.get("tool.loop.concurrency", None),  # type: ignore
            timeout=self.settings.get("tool.loop.timeout", None),  # type: ignore
        )

        if len(records) > 0:
            prompt.set("action_results", self.tool.to_action_results(records))
            prompt.set("extra_instruction", self.tool.TOOL_RESULT_QUOTE_NOTICE)
            self.__tool_logs = records

    async def __broadcast_prefix(self, full_result_data: "AgentlyModelResult", _):
        if len(self.__tool_logs) == 0:
            return

        for tool_log in self.__tool_logs:
            yield "tool", tool_log

        if "extra" not in full_result_data:
            full_result_data["extra"] = {}
        if isinstance(full_result_data["extra"], dict) and "tool_logs" not in full_result_data["extra"]:
            full_result_data["extra"]["tool_logs"] = []
        if (
            "extra" in full_result_data
            and isinstance(full_result_data["extra"], dict)
            and "tool_logs" in full_result_data["extra"]
            and isinstance(full_result_data["extra"]["tool_logs"], list)
        ):
            full_result_data["extra"]["tool_logs"].extend(self.__tool_logs)
        if self.settings.get("runtime.show_tool_logs"):
            for tool_log in self.__tool_logs:
                await self.async_system_message(
                    "TOOL",
                    "\n".join(
                        [f"{ key }: { value }" for key, value in tool_log.items()],
                    ),
                )
        self.__tool_logs = []



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import asyncio
import inspect
import warnings

from typing import Any, AsyncGenerator, Literal, TYPE_CHECKING, cast, overload, Generator

from agently.utils import FunctionShifter, DataLocator

if TYPE_CHECKING:
    from pydantic import BaseModel

    from agently.core import Prompt, ExtensionHandlers
    from agently.core.PluginManager import PluginManager
    from agently.utils import Settings
    from agently.types.data import (
        AgentlyModelResponseMessage,
        InstantStreamingContentType,
        ResponseContentType,
        SpecificEvents,
        StreamingData,
    )
    from agently.types.plugins import ResponseParser


DEFAULT_SPECIFIC_EVENTS: "SpecificEvents" = [
    "reasoning_delta",
    "delta",
    "reasoning_done",
    "done",
    "tool_calls",
]


class ModelResponseResult:
    def __init__(
        self,
        agent_name: str,
        response_id: str,
        prompt: "Prompt",
        response_generator: AsyncGenerator["AgentlyModelResponseMessage", None],
        plugin_manager: "PluginManager",
        settings: "Settings",
        extension_handlers: "ExtensionHandlers",
    ):
        self.agent_name = agent_name
        self.plugin_manager = plugin_manager
        self.settings = settings
        ResponseParser = cast(
            type["ResponseParser"],
            self.plugin_manager.get_plugin(
                "ResponseParser",
                str(self.settings["plugins.ResponseParser.activate"]),
            ),
        )
        self._response_id = response_id
        self._extension_handlers = extension_handlers
        self._response_parser = ResponseParser(agent_name, response_id, prompt, response_generator, self.settings)
        self._finally_handlers_ran = False
        self._finally_handlers_lock = asyncio.Lock()
        self._run_finally_handlers_once_sync = FunctionShifter.syncify(self._run_finally_handlers_once)
        self.prompt = prompt
        self.full_result_data = self._response_parser.full_result_data
        self.get_meta = FunctionShifter.syncify(self.async_get_meta)
        self.get_text = FunctionShifter.syncify(self.async_get_text)
        self.get_data = FunctionShifter.syncify(self.async_get_data)
        self.get_data_object = FunctionShifter.syncify(self.async_get_data_object)

    async def _run_finally_handlers_once(self):
        if self._finally_handlers_ran:
            return
        async with self._finally_handlers_lock:
            if self._finally_handlers_ran:
                return
            # Mark as executed before invoking handlers so handlers can safely
            # call result getters without re-entering this hook chain.
            self._finally_handlers_ran = True
            finally_handlers = self._extension_handlers.get("finally", [])
            for handler in finally_handlers:
                if inspect.iscoroutinefunction(handler):
                    await handler(
                        self,
                        self.settings,
                    )
                elif inspect.isgeneratorfunction(handler):
                    for _ in handler(
                        self,
                        self.settings,
                    ):
                        pass
                elif inspect.isasyncgenfunction(handler):
                    async for _ in handler(
                        self,
                        self.settings,
                    ):
                        pass
                elif inspect.isfunction(handler):
                    handler(
                        self,
                        self.settings,
                    )

    @overload
    async def async_get_data(
        self,
        *,
        type: Literal['parsed'],
        ensure_keys: list[str],
        key_style: Literal["dot", "slash"] = "dot",
        max_retries: int = 3,
        raise_ensure_failure: bool = True,
        _retry_count: int = 0,
    ) -> dict[str, Any]: ...

    @overload
    async def async_get_data(
        self,
        *,
        type: Literal['original', 'parsed', 'all'] = "parsed",
        ensure_keys: list[str] | None = None,
        key_style: Literal["dot", "slash"] = "dot",
        max_retries: int = 3,
        raise_ensure_failure: bool = True,
        _retry_count: int = 0,
    ) -> Any: ...

    async def async_get_data(
        self,
        *,
        type: Literal['original', 'parsed', 'all'] = "parsed",
        ensure_keys: list[str] | None = None,
        key_style: Literal["dot", "slash"] = "dot",
        max_retries: int = 3,
        raise_ensure_failure: bool = True,
        _retry_count: int = 0,
    ) -> Any:
        if type == "parsed" and ensure_keys:
            try:
                data = await self._response_parser.async_get_data(type=type)
                for ensure_key in ensure_keys:
                    EMPTY = object()
                    if DataLocator.locate_path_in_dict(data, ensure_key, key_style, default=EMPTY) is EMPTY:
                        raise
                await self._run_finally_handlers_once()
                return data
            except:
                from agently.base import async_system_message
                from agently.core.ModelResponse import ModelResponse

                await async_system_message(
                    "MODEL_REQUEST",
                    {
                        "agent_name": self.agent_name,
                        "response_id": self._response_id,
                        "content": {
                            "stage": "No Target Data in Response, Preparing Retry",
                            "detail": f"\n[Response]: { await self._response_parser.async_get_text() }\n"
                            f"[Retried Times]: { _retry_count }",
                        },
                    },
                    self.settings,
                )

                if _retry_count < max_retries:
                    data = await ModelResponse(
                        self.agent_name,
                        self.plugin_manager,
                        self.settings,
                        self.prompt,
                        self._extension_handlers,
                    ).result.async_get_data(
                        type=type,
                        ensure_keys=ensure_keys,
                        key_style=key_style,
                        max_retries=max_retries,
                        raise_ensure_failure=raise_ensure_failure,
                        _retry_count=_retry_count + 1,
                    )
                    await self._run_finally_handlers_once()
                    return data
                else:
                    if raise_ensure_failure:
                        await self._run_finally_handlers_once()
                        raise ValueError(
                            f"Can not generate ensure keys { ensure_keys } within { max_retries } retires."
                        )
                    data = await self._response_parser.async_get_data(type=type)
                    await self._run_finally_handlers_once()
                    return data
        data = await self._response_parser.async_get_data(type=type)
        await self._run_finally_handlers_once()
        return data

    @overload
    async def async_get_data_object(
        self,
    ) -> "BaseModel | None": ...

    @overload
    async def async_get_data_object(
        self,
        *,
        ensure_keys: list[str],
        key_style: Literal["dot", "slash"] = "dot",
        max_retries: int = 3,
        raise_ensure_failure: bool = True,
    ) -> "BaseModel": ...

    @overload
    async def async_get_data_object(
        self,
        *,
        ensure_keys: None,
        key_style: Literal["dot", "slash"] = "dot",
        max_retries: int = 3,
        raise_ensure_failure: bool = True,
    ) -> "BaseModel | None": ...

    async def async_get_data_object(
        self,
        *,
        ensure_keys: list[str] | None = None,
        key_style: Literal["dot", "slash"] = "dot",
        max_retries: int = 3,
        raise_ensure_failure: bool = True,
    ):
        if ensure_keys:
            await self.async_get_data(
                ensure_keys=ensure_keys,
                key_style=key_style,
                max_retries=max_retries,
                _retry_count=0,
                raise_ensure_failure=raise_ensure_failure,
            )
            result_object = await self._response_parser.async_get_data_object()
            await self._run_finally_handlers_once()
            return result_object
        result_object = await self._response_parser.async_get_data_object()
        await self._run_finally_handlers_once()
        return result_object

    async def async_get_meta(self):
        meta = await self._response_parser.async_get_meta()
        await self._run_finally_handlers_once()
        return meta

    async def async_get_text(self):
        text = await self._response_parser.async_get_text()
        await self._run_finally_handlers_once()
        return text

    @overload
    def get_generator(
        self,
        type: "InstantStreamingContentType",
        *,
        specific: "SpecificEvents" = DEFAULT_SPECIFIC_EVENTS,
    ) -> Generator["StreamingData", None, None]: ...

    @overload
    def get_generator(
        self,
        type: Literal["all"],
        *,
        specific: "SpecificEvents" = DEFAULT_SPECIFIC_EVENTS,
    ) -> Generator[tuple[str, Any], None, None]: ...

    @overload
    def get_generator(
        self,
        type: Literal["delta", "specific", "original"],
        *,
        specific: "SpecificEvents" = DEFAULT_SPECIFIC_EVENTS,
    ) -> Generator[str, None, None]: ...

    @overload
    def get_generator(
        self,
        type: "ResponseContentType | None" = "delta",
        *,
        specific: "SpecificEvents" = DEFAULT_SPECIFIC_EVENTS,
    ) -> Generator: ...

    def get_generator(
        self,
        type: "ResponseContentType | None" = None,
        content: "ResponseContentType | None" = None,
        *,
        specific: "SpecificEvents" = DEFAULT_SPECIFIC_EVENTS,
    ) -> Generator:
        if type is None:
            if content is not None:
                warnings.warn(
                    "Parameter `content` in method .get_generator() is  deprecated and will be removed in future "
                    "version, please use parameter `type` instead."
                )
                type = content
            else:
                type = "delta"
        parsed_generator = self._response_parser.get_generator(type=type, specific=specific)
        completed = False
        for data in parsed_generator:
            yield data
        completed = True
        if completed:
            self._run_finally_handlers_once_sync()

    @overload
    def get_async_generator(
        self,
        type: "InstantStreamingContentType",
        *,
        specific: "SpecificEvents" = DEFAULT_SPECIFIC_EVENTS,
    ) -> AsyncGenerator["StreamingData", None]: ...

    @overload
    def get_async_generator(
        self,
        type: Literal["all"],
        *,
        specific: "SpecificEvents" = DEFAULT_SPECIFIC_EVENTS,
    ) -> AsyncGenerator[tuple[str, Any], None]: ...

    @overload
    def get_async_generator(
        self,
        type: Literal["delta", "specific", "original"],
        *,
        specific: "SpecificEvents" = DEFAULT_SPECIFIC_EVENTS,
    ) -> AsyncGenerator[str, None]: ...

    @overload
    def get_async_generator(
        self,
        type: "ResponseContentType | None" = "delta",
        *,
        specific: "SpecificEvents" = DEFAULT_SPECIFIC_EVENTS,
    ) -> AsyncGenerator: ...

    async def get_async_generator(
        self,
        type: "ResponseContentType | None" = None,
        content: "ResponseContentType | None" = None,
        *,
        specific: "SpecificEvents" = DEFAULT_SPECIFIC_EVENTS,
    ) -> AsyncGenerator:
        if type is None:
            if content is not None:
                warnings.warn(
                    "Parameter `content` in method .get_async_generator() is  deprecated and will be removed in "
                    "future version, please use parameter `type` instead."
                )
                type = content
            else:
                type = "delta"
        parsed_generator = self._response_parser.get_async_generator(type=type, specific=specific)
        completed = False
        async for data in parsed_generator:
            yield data
        completed = True
        if completed:
            await self._run_finally_handlers_once()



# Copyright 2023-2026 AgentEra(Agently.Tech)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Mapping, Sequence, Any, Literal, get_origin, get_args
from collections import OrderedDict


class DataPathBuilder:
    @staticmethod
    def build_dot_path(keys: list[str | int]) -> str:
        if not keys:
            return ""

        parts = []
        for i, key in enumerate(keys):
            if isinstance(key, int):
                parts.append(f"[{ key }]")
            elif str(key) in ("*", "[]", "[*]"):
                parts.append("[*]")
            else:
                if i == 0:
                    parts.append(str(key))
                else:
                    parts.append(f".{key}")
        return "".join(parts)

    @staticmethod
    def build_slash_path(keys: list[str | int]) -> str:
        if not keys:
            return ""

        parts = ["/"]
        for key in keys:
            parts.append(str(key))
            parts.append("/")
        return "".join(parts[:-1])

    @staticmethod
    def convert_dot_to_slash(dot_path: str) -> str:
        """
        Convert dot-style path to slash-style path.
        Example: 'user.name' -> '/user/name', 'tasks[*].id' -> '/tasks/[*]/id'
        """
        if not dot_path:
            return "/"

        parts = []
        buffer = ""
        i = 0
        while i < len(dot_path):
            if dot_path[i] == "[":
                if buffer:
                    parts.append(buffer)
                    buffer = ""
                end = dot_path.find("]", i)
                parts.append(dot_path[i : end + 1])
                i = end + 1
            elif dot_path[i] == ".":
                if buffer:
                    parts.append(buffer)
                    buffer = ""
                i += 1
            else:
                buffer += dot_path[i]
                i += 1
        if buffer:
            parts.append(buffer)

        return "/" + "/".join(parts)

    @staticmethod
    def convert_slash_to_dot(slash_path: str) -> str:
        """
        Convert slash-style path to dot-style path.
        Example: '/user/name' -> 'user.name', '/tasks/[*]/id' -> 'tasks[*].id'
        """
        if not slash_path or slash_path == "/":
            return ""

        parts = slash_path.strip("/").split("/")
        dot_parts = []
        for part in parts:
            if part.startswith("[") and part.endswith("]"):
                dot_parts.append(part)
            elif not dot_parts:
                dot_parts.append(part)
            else:
                dot_parts.append(f".{part}")

        return "".join(dot_parts)

    @staticmethod
    def extract_possible_paths(agently_output_dict: dict[str, Any], *, style: Literal["dot", "slash"] = "dot"):
        """
        Extract Agently style output dictionary to find all possible paths from model structured response.
        """
        if not isinstance(agently_output_dict, dict):
            raise TypeError("extract_all_paths expects a dict[str, Any] as input")
        all_paths: set[str] = set()

        def extract_paths(value: Any, path_keys: list[str | int]):
            current_path = (
                DataPathBuilder.build_dot_path(path_keys)
                if style == "dot"
                else DataPathBuilder.build_slash_path(path_keys)
            )
            all_paths.add(current_path)

            if isinstance(value, tuple) and value:
                extract_paths(value[0], path_keys)
                return

            if isinstance(value, dict):
                for k, v in value.items():
                    extract_paths(v, path_keys + [k])
                return

            origin = get_origin(value)
            if origin is list:
                args = get_args(value)
                if args:
                    extract_paths(args[0], path_keys + ["[*]"])
                return

            if isinstance(value, list):
                for idx, item in enumerate(value):
                    extract_paths(item, path_keys + ["[*]"])
                return

        extract_paths(agently_output_dict, [])
        return all_paths

    @staticmethod
    def extract_parsing_key_orders(
        agently_output_dict: dict[str, Any], *, style: Literal["dot", "slash"] = "dot"
    ) -> list[str]:
        """
        Traverse the schema in recursive order and return a list of paths preserving the field definition order.
        """
        ordered_paths: OrderedDict[str, None] = OrderedDict()

        def traverse(value: Any, path_keys: list[str | int]):
            if isinstance(value, tuple) and value:
                traverse(value[0], path_keys)
                current_path = (
                    DataPathBuilder.build_dot_path(path_keys)
                    if style == "dot"
                    else DataPathBuilder.build_slash_path(path_keys)
                )
                ordered_paths[current_path] = None
                return

            if isinstance(value, dict):
                for k, v in value.items():
                    traverse(v, path_keys + [k])
                current_path = (
                    DataPathBuilder.build_dot_path(path_keys)
                    if style == "dot"
                    else DataPathBuilder.build_slash_path(path_keys)
                )
                ordered_paths[current_path] = None
                return

            origin = get_origin(value)
            if origin is list:
                args = get_args(value)
                if args:
                    traverse(args[0], path_keys + ["[*]"])
                current_path = (
                    DataPathBuilder.build_dot_path(path_keys)
                    if style == "dot"
                    else DataPathBuilder.build_slash_path(path_keys)
                )
                ordered_paths[current_path] = None
                return

            if isinstance(value, list):
                for item in value:
                    traverse(item, path_keys + ["[*]"])
                current_path = (
                    DataPathBuilder.build_dot_path(path_keys)
                    if style == "dot"
                    else DataPathBuilder.build_slash_path(path_keys)
                )
                ordered_paths[current_path] = None
                return

            current_path = (
                DataPathBuilder.build_dot_path(path_keys)
                if style == "dot"
                else DataPathBuilder.build_slash_path(path_keys)
            )
            ordered_paths[current_path] = None

        traverse(agently_output_dict, [])
        return list(ordered_paths.keys())

    @staticmethod
    def get_value_by_path(
        data: Mapping[str, Any] | Sequence[Any], path: str, *, style: Literal["dot", "slash"] = "dot"
    ) -> Any:
        """
        Retrieve the value from a dictionary based on the given path.
        If the path does not exist, return None.

        :param data: The dictionary to search.
        :param path: The path to the desired value (dot or slash style).
        :param style: The style of the path, either "dot" or "slash".
        :return: The value at the specified path, or None if not found.
        """
        if not isinstance(data, dict):
            raise TypeError("The data parameter must be a dictionary.")

        if not path:
            return None

        keys = []
        if style == "dot":
            buffer = ""
            i = 0
            while i < len(path):
                if path[i] == "[":
                    if buffer:
                        keys.append(buffer)
                        buffer = ""
                    end = path.find("]", i)
                    keys.append(path[i + 1 : end])  # Keep as string to handle [*]
                    i = end + 1
                elif path[i] == ".":
                    if buffer:
                        keys.append(buffer)
                        buffer = ""
                    i += 1
                else:
                    buffer += path[i]
                    i += 1
            if buffer:
                keys.append(buffer)
        elif style == "slash":
            keys: list[Any] = path.strip("/").split("/")
            for i, key in enumerate(keys):
                if key.isdigit():
                    keys[i] = int(key)

        def resolve(current: Any, remaining_keys: list[Any]) -> Any:
            if not remaining_keys:
                return current

            key = remaining_keys[0]
            if key == "[*]":
                if isinstance(current, Sequence) and not isinstance(current, str):
                    results = []
                    for item in current:
                        result = resolve(item, remaining_keys[1:])
                        if isinstance(result, list):
                            results.extend(result)
                        elif result is not None:
                            results.append(result)
                    return results
                else:
                    return None
            elif isinstance(current, Mapping) and key in current:
                return resolve(current[key], remaining_keys[1:])
            elif (
                not isinstance(current, str)
                and isinstance(current, Sequence)
                and isinstance(key, int)
                and 0 <= key < len(current)
            ):
                return resolve(current[key], remaining_keys[1:])
            else:
                return None

        return resolve(data, keys)
